]> git.sesse.net Git - stockfish/blobdiff - src/tt.h
Better code for hash table generation
[stockfish] / src / tt.h
index 1caa277e821ac0463110fe0df6487b74f7e20f79..a750b6c465706c1daa22947b082b42a8decbf641 100644 (file)
--- a/src/tt.h
+++ b/src/tt.h
@@ -1,7 +1,6 @@
 /*
   Stockfish, a UCI chess playing engine derived from Glaurung 2.1
-  Copyright (C) 2004-2008 Tord Romstad (Glaurung author)
-  Copyright (C) 2008-2013 Marco Costalba, Joona Kiiski, Tord Romstad
+  Copyright (C) 2004-2021 The Stockfish developers (see AUTHORS file)
 
   Stockfish is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-#if !defined(TT_H_INCLUDED)
+#ifndef TT_H_INCLUDED
 #define TT_H_INCLUDED
 
 #include "misc.h"
 #include "types.h"
 
-/// The TTEntry is the class of transposition table entries
+/// TTEntry struct is the 10 bytes transposition table entry, defined as below:
 ///
-/// A TTEntry needs 128 bits to be stored
-///
-/// bit  0-31: key
-/// bit 32-63: data
-/// bit 64-79: value
-/// bit 80-95: depth
-/// bit 96-111: static value
-/// bit 112-127: margin of static value
-///
-/// the 32 bits of the data field are so defined
-///
-/// bit  0-15: move
-/// bit 16-20: not used
-/// bit 21-22: value type
-/// bit 23-31: generation
-
-class TTEntry {
-
-public:
-  void save(uint32_t k, Value v, Bound b, Depth d, Move m, int g, Value ev, Value em) {
-
-    key32        = (uint32_t)k;
-    move16       = (uint16_t)m;
-    bound        = (uint8_t)b;
-    generation8  = (uint8_t)g;
-    value16      = (int16_t)v;
-    depth16      = (int16_t)d;
-    evalValue    = (int16_t)ev;
-    evalMargin   = (int16_t)em;
-  }
-  void set_generation(int g) { generation8 = (uint8_t)g; }
-
-  uint32_t key() const      { return key32; }
-  Depth depth() const       { return (Depth)depth16; }
-  Move move() const         { return (Move)move16; }
-  Value value() const       { return (Value)value16; }
-  Bound type() const        { return (Bound)bound; }
-  int generation() const    { return (int)generation8; }
-  Value eval_value() const  { return (Value)evalValue; }
-  Value eval_margin() const { return (Value)evalMargin; }
+/// key        16 bit
+/// depth       8 bit
+/// generation  5 bit
+/// pv node     1 bit
+/// bound type  2 bit
+/// move       16 bit
+/// value      16 bit
+/// eval value 16 bit
+
+struct TTEntry {
+
+  Move  move()  const { return (Move )move16; }
+  Value value() const { return (Value)value16; }
+  Value eval()  const { return (Value)eval16; }
+  Depth depth() const { return (Depth)depth8 + DEPTH_OFFSET; }
+  bool is_pv()  const { return (bool)(genBound8 & 0x4); }
+  Bound bound() const { return (Bound)(genBound8 & 0x3); }
+  void save(Key k, Value v, bool pv, Bound b, Depth d, Move m, Value ev);
 
 private:
-  uint32_t key32;
+  friend class TranspositionTable;
+
+  uint16_t key16;
+  uint8_t  depth8;
+  uint8_t  genBound8;
   uint16_t move16;
-  uint8_t bound, generation8;
-  int16_t value16, depth16, evalValue, evalMargin;
+  int16_t  value16;
+  int16_t  eval16;
 };
 
 
/// A TranspositionTable consists of a power of 2 number of clusters and each
/// cluster consists of ClusterSize number of TTEntry. Each non-empty entry
/// contains information of exactly one position. Size of a cluster shall not be
/// bigger than a cache line size. In case it is less, it should be padded to
/// guarantee always aligned accesses.
 
 class TranspositionTable {
 
-  static const unsigned ClusterSize = 4; // A cluster is 64 Bytes
-
-public:
- ~TranspositionTable() { free(mem); }
-  void new_search() { generation++; }
-
-  TTEntry* probe(const Key key) const;
-  TTEntry* first_entry(const Key key) const;
-  void refresh(const TTEntry* tte) const;
-  void set_size(size_t mbSize);
-  void clear();
-  void store(const Key key, Value v, Bound type, Depth d, Move m, Value statV, Value kingD);
-
-private:
-  uint32_t hashMask;
-  TTEntry* table;
-  void* mem;
-  uint8_t generation; // Size must be not bigger then TTEntry::generation8
-};
-
-extern TranspositionTable TT;
+  static constexpr int ClusterSize = 3;
 
+  struct Cluster {
+    TTEntry entry[ClusterSize];
+    char padding[2]; // Pad to 32 bytes
+  };
 
-/// TranspositionTable::first_entry() returns a pointer to the first entry of
-/// a cluster given a position. The lowest order bits of the key are used to
-/// get the index of the cluster.
+  static_assert(sizeof(Cluster) == 32, "Unexpected Cluster size");
 
-inline TTEntry* TranspositionTable::first_entry(const Key key) const {
+  // Constants used to refresh the hash table periodically
+  static constexpr unsigned GENERATION_BITS  = 3;                                // nb of bits reserved for other things
+  static constexpr int      GENERATION_DELTA = (1 << GENERATION_BITS);           // increment for generation field
+  static constexpr int      GENERATION_CYCLE = 255 + (1 << GENERATION_BITS);     // cycle length
+  static constexpr int      GENERATION_MASK  = (0xFF << GENERATION_BITS) & 0xFF; // mask to pull out generation number
 
-  return table + ((uint32_t)key & hashMask);
-}
+public:
+ ~TranspositionTable() { aligned_large_pages_free(table); }
+  void new_search() { generation8 += GENERATION_DELTA; } // Lower bits are used for other things
+  TTEntry* probe(const Key key, bool& found) const;
+  int hashfull() const;
+  void resize(size_t mbSize);
+  void clear();
 
+  TTEntry* first_entry(const Key key) const {
+    return &table[mul_hi64(key, clusterCount)].entry[0];
+  }
 
-/// TranspositionTable::refresh() updates the 'generation' value of the TTEntry
-/// to avoid aging. Normally called after a TT hit.
+private:
+  friend struct TTEntry;
 
-inline void TranspositionTable::refresh(const TTEntry* tte) const {
+  size_t clusterCount;
+  Cluster* table;
+  uint8_t generation8; // Size must be not bigger than TTEntry::genBound8
+};
 
-  const_cast<TTEntry*>(tte)->set_generation(generation);
-}
+extern TranspositionTable TT;
 
-#endif // !defined(TT_H_INCLUDED)
+#endif // #ifndef TT_H_INCLUDED
@@ -1,7 +1,6 @@
 
 
-/// A TranspositionTable consists of a power of 2 number of clusters and each
-/// cluster consists of ClusterSize number of TTEntry. Each non-empty entry
-/// contains information of exactly one position. Size of a cluster shall not be
-/// bigger than a cache line size. In case it is less, it should be padded to
-/// guarantee always aligned accesses.
+/// A TranspositionTable is an array of Cluster, of size clusterCount. Each
+/// cluster consists of ClusterSize number of TTEntry. Each non-empty TTEntry
+/// contains information on exactly one position. The size of a Cluster should
+/// divide the size of a cache line for best performance, as the cacheline is
+/// prefetched when possible.
 
 class TranspositionTable {